@@ -17,7 +17,8 @@
 **  y is overwriten with b
 **
 **  If flag=0, solve Ab=y,
-**         =1, solve L sqrt(D)b=y, where A=LDL', L lower triangular
+**         =1, solve sqrt(D)L' b=y, where A=LDL', L lower triangular
+**         =2, solve L sqrt(D) b=y
 **  Terry Therneau
 */
 #include "bdsmatrix.h"
@@ -35,39 +36,38 @@ void chsolve4(double **matrix, int n, int nblock, int *bsize,
     n2 = n-m;
 
     /*
-    ** solve Fz =y , using eq 2.2.2 of A. George and A Liu, Computer Solution
+    ** solve Lz =y , using eq 2.2.2 of A. George and A Liu, Computer Solution
     **  of Large Sparse Positive Definite Systems, Prentice-Hall, 1981.
     **
     **  first the block diagonal portion
     */
-    i =0;
-    ii =0;
-    for (block=0; block<nblock; block++) {
-        /* process block i */
-        for (j=bsize[block]; j>0; j--) {
-            temp = y[i];
-            for (k=1; k<j; k++) 
-		y[i+k] -= bd[ii+k]*temp;
-            for (k=0; k<n2; k++)
-		y[m+k] -= matrix[k][i] *temp;
-            i++;
-            ii += j;
+    if (flag <2) {
+	i =0;
+	ii =0;
+	for (block=0; block<nblock; block++) {
+	    /* process block i */
+	    for (j=bsize[block]; j>0; j--) {
+		temp = y[i];
+		for (k=1; k<j; k++) 
+		    y[i+k] -= bd[ii+k]*temp;
+		for (k=0; k<n2; k++)
+		    y[m+k] -= matrix[k][i] *temp;
+		i++;
+		ii += j;
             }
         }
+	/*
+	** Finish up by doing the dense portion of L'z=y
+	*/
+	for (j=0; j<n2; j++) {
+	    temp = y[j+m];
+	    for (k=j+1; k<n2; k++) y[k+m] -= temp * matrix[k][j+m];
+	}
+    }		
     /*
-    ** At this point i and ii point to the END of the bd array, and need to
-    **  be saved for later use
-    ** Finish up by doing the dense portion of Fz=y
+    ** Now the diagonal
     */
-    for (j=0; j<n2; j++) {
-         temp = y[j+m];
-         for (k=j+1; k<n2; k++) y[k+m] -= temp * matrix[k][j+m];
-         }
-
-    /*
-    ** Now the second half of the work
-    */
-    if (flag==1) {
+    if (flag> 0) {
         /* solve sqrt(D) b =z */
         i =0;
         ii=0;
@@ -88,7 +88,28 @@ void chsolve4(double **matrix, int n, int nblock, int *bsize,
             }
         }
 
-    else {
+    else { /* solve Db =z */
+        i =0;
+        ii=0;
+        for (block=0; block<nblock; block++) {
+            for (j=bsize[block]; j>0; j--) {
+		/* ii points to A[i,i] */
+		if (bd[ii] >0) y[i] /= bd[ii];
+		else           y[i] = 0;
+		i++;
+		ii += j;
+	        }
+            }
+        /* dense portion */
+        for (j=0; j<n2; j++) {
+            temp = matrix[j][j+i];
+            if (temp>0) y[j+i] /= temp;
+            else y[i+j] = 0;
+            }
+        }
+
+    /* solve L'b =z */
+    if (flag != 1) {
         /*
         ** solve DF'b =z, using equation 2.2.1
         */
@@ -96,7 +117,7 @@ void chsolve4(double **matrix, int n, int nblock, int *bsize,
         for (j=(n2-1); j>=0; j--) {
        	  if (matrix[j][j+m]==0)  y[j+m] =0;
        	  else {
-       	      temp = y[j+m]/matrix[j][j+m];
+       	      temp = y[j+m];
        	      for (k= j+1; k<n2; k++)
        		   temp -= y[k+m]*matrix[k][j+m];
        	      y[j+m] = temp;
@@ -109,7 +130,7 @@ void chsolve4(double **matrix, int n, int nblock, int *bsize,
        	     ii -= blocksize;
        	     if (bd[ii] ==0) y[i] =0;
        	     else {
-       		 temp = y[i] / bd[ii];
+       		 temp = y[i];
        		 for (j=1; j<blocksize; j++)
        		     temp -= y[i+j]*bd[ii+j];
        		 for (j=0; j<n2; j++)
